@@ -27,6 +27,11 @@ from util import (wait_for_mount, umount, cleanup, base_cmdline,
                   fuse_proto, fuse_caps, powerset, parse_kernel_version)
 from os.path import join as pjoin
 import logging
+from enum import Enum
+
+class InodeCheck(Enum):
+    EXACT = 1
+    NONZERO = 2
 
 pytestmark = fuse_test_marker()
 
@@ -138,7 +143,7 @@ def test_hello(tmpdir, name, options, cmdline_builder, output_checker):
 
 @pytest.mark.parametrize("writeback", (False, True))
 @pytest.mark.parametrize("name", ('passthrough', 'passthrough_plus',
-                           'passthrough_fh', 'passthrough_ll'))
+                           'passthrough_fh', 'passthrough_ll', 'passthrough_zero_ino'))
 @pytest.mark.parametrize("debug", (False, True))
 def test_passthrough(short_tmpdir, name, debug, output_checker, writeback):
     # Avoid false positives from libfuse debug messages
@@ -153,14 +158,30 @@ def test_passthrough(short_tmpdir, name, debug, output_checker, writeback):
     mnt_dir = str(short_tmpdir.mkdir('mnt'))
     src_dir = str(short_tmpdir.mkdir('src'))
 
+    inode_check = InodeCheck.EXACT
     if name == 'passthrough_plus':
         cmdline = base_cmdline + \
                   [ pjoin(basename, 'example', 'passthrough'),
                     '--plus', '-f', mnt_dir ]
-    else:
+    elif name == 'passthrough_zero_ino':
+        cmdline = base_cmdline + \
+                  [ pjoin(basename, 'example', 'passthrough'),
+                    '--plus', '--readdir-zero-inodes', '-f', mnt_dir ]
+        inode_check = InodeCheck.NONZERO
+    elif name == 'passthrough_ll':
+        cmdline = base_cmdline + \
+                  [ pjoin(basename, 'example', name),
+                    '-f', mnt_dir, '-o', 'timeout=0' ]
+    else:  # passthrough and passthrough_fh
         cmdline = base_cmdline + \
                   [ pjoin(basename, 'example', name),
                     '-f', mnt_dir ]
+
+    # Set all timeouts to 0 for everything except passthrough_ll
+    # (this includes passthrough, passthrough_plus, and passthrough_fh)
+    if name != 'passthrough_ll':
+        cmdline.extend(['-o', 'entry_timeout=0,negative_timeout=0,attr_timeout=0,ac_attr_timeout=0'])
+
     if debug:
         cmdline.append('-d')
 
@@ -169,7 +190,9 @@ def test_passthrough(short_tmpdir, name, debug, output_checker, writeback):
             pytest.skip('example does not support writeback caching')
         cmdline.append('-o')
         cmdline.append('writeback')
-        
+
+    print(f"\nDebug: Command line: {' '.join(cmdline)}")
+
     mount_process = subprocess.Popen(cmdline, stdout=output_checker.fd,
                                      stderr=output_checker.fd)
     try:
@@ -177,12 +200,12 @@ def test_passthrough(short_tmpdir, name, debug, output_checker, writeback):
         work_dir = mnt_dir + src_dir
 
         tst_statvfs(work_dir)
-        tst_readdir(src_dir, work_dir)
-        tst_readdir_big(src_dir, work_dir)
+        tst_readdir(src_dir, work_dir, inode_check)
+        tst_readdir_big(src_dir, work_dir, inode_check)
         tst_open_read(src_dir, work_dir)
         tst_open_write(src_dir, work_dir)
         tst_create(work_dir)
-        tst_passthrough(src_dir, work_dir)
+        tst_passthrough(src_dir, work_dir, inode_check)
         tst_append(src_dir, work_dir)
         tst_seek(src_dir, work_dir)
         tst_mkdir(work_dir)
@@ -195,7 +218,8 @@ def test_passthrough(short_tmpdir, name, debug, output_checker, writeback):
         # Underlying fs may not have full nanosecond resolution
         tst_utimens(work_dir, ns_tol=1000)
 
-        tst_link(work_dir)
+        if inode_check == InodeCheck.EXACT:
+            tst_link(work_dir)
         tst_truncate_path(work_dir)
         tst_truncate_fd(work_dir)
         tst_open_unlink(work_dir)
@@ -364,7 +388,8 @@ def test_null(tmpdir, output_checker):
 
 @pytest.mark.skipif(fuse_proto < (7,12),
                     reason='not supported by running kernel')
-@pytest.mark.parametrize("only_expire", ("invalidate_entries", "expire_entries"))
+@pytest.mark.parametrize("only_expire", ("invalidate_entries",
+                                         "expire_entries", "inc_epoch"))
 @pytest.mark.parametrize("notify", (True, False))
 def test_notify_inval_entry(tmpdir, only_expire, notify, output_checker):
     mnt_dir = str(tmpdir)
@@ -378,6 +403,10 @@ def test_notify_inval_entry(tmpdir, only_expire, notify, output_checker):
         cmdline.append('--only-expire')
         if "FUSE_CAP_EXPIRE_ONLY" not in fuse_caps:
             pytest.skip('only-expire not supported by running kernel')
+    elif only_expire == "inc_epoch":
+        cmdline.append('--inc-epoch')
+        if fuse_proto < (7,44):
+            pytest.skip('inc-epoch not supported by running kernel')
     mount_process = subprocess.Popen(cmdline, stdout=output_checker.fd,
                                      stderr=output_checker.fd)
     try:
@@ -740,7 +769,17 @@ def tst_link(mnt_dir):
 
     os.unlink(name1)
 
-def tst_readdir(src_dir, mnt_dir):
+def tst_inodes_nonzero(lines):
+    inode_nums = [int(line.split()[0]) for line in lines]
+    assert all(i != 0 for i in inode_nums), inode_nums
+
+def tst_inode(inode_check, actual, expected):
+    if inode_check == InodeCheck.EXACT:
+        assert expected == actual
+    elif inode_check == InodeCheck.NONZERO:
+        assert actual != 0
+
+def tst_readdir(src_dir, mnt_dir, inode_check=InodeCheck.EXACT):
     newdir = name_generator()
 
     src_newdir = pjoin(src_dir, newdir)
@@ -761,16 +800,18 @@ def tst_readdir(src_dir, mnt_dir):
     assert listdir_is == listdir_should
 
     inodes_is = readdir_inode(mnt_newdir)
-    inodes_should = readdir_inode(src_newdir)
-    assert inodes_is == inodes_should
+    if inode_check == InodeCheck.EXACT:
+        inodes_should = readdir_inode(src_newdir)
+        assert inodes_is == inodes_should
+    elif inode_check == InodeCheck.NONZERO:
+        tst_inodes_nonzero(inodes_is)
 
     os.unlink(file_)
     os.unlink(subfile)
     os.rmdir(subdir)
     os.rmdir(src_newdir)
 
-def tst_readdir_big(src_dir, mnt_dir):
-
+def tst_readdir_big(src_dir, mnt_dir, inode_check=InodeCheck.EXACT):
     # Add enough entries so that readdir needs to be called
     # multiple times.
     fnames = []
@@ -786,13 +827,17 @@ def tst_readdir_big(src_dir, mnt_dir):
     assert listdir_is == listdir_should
 
     inodes_is = readdir_inode(mnt_dir)
-    inodes_should = readdir_inode(src_dir)
-    assert inodes_is == inodes_should
+    if inode_check == InodeCheck.EXACT:
+        inodes_should = readdir_inode(src_dir)
+        assert inodes_is == inodes_should
+    elif inode_check == InodeCheck.NONZERO:
+        tst_inodes_nonzero(inodes_is)
 
     for fname in fnames:
+        # A comment just to get a diff
         stat_src = os.stat(pjoin(src_dir, fname))
         stat_mnt = os.stat(pjoin(mnt_dir, fname))
-        assert stat_src.st_ino == stat_mnt.st_ino
+        tst_inode(inode_check, stat_mnt.st_ino, stat_src.st_ino)
         assert stat_src.st_mtime == stat_mnt.st_mtime
         assert stat_src.st_ctime == stat_mnt.st_ctime
         assert stat_src.st_size == stat_mnt.st_size
@@ -868,28 +913,61 @@ def tst_utimens(mnt_dir, ns_tol=0):
         assert abs(fstat.st_atime_ns - atime_ns) <= ns_tol
         assert abs(fstat.st_mtime_ns - mtime_ns) <= ns_tol
 
-def tst_passthrough(src_dir, mnt_dir):
+def tst_passthrough(src_dir, mnt_dir, inode_check=InodeCheck.EXACT):
     name = name_generator()
     src_name = pjoin(src_dir, name)
-    mnt_name = pjoin(src_dir, name)
+    mnt_name = pjoin(mnt_dir, name)
+
+    print(f"\nDebug: Creating file {name}")
+    print(f"Debug: src_name={src_name}")
+    print(f"Debug: mnt_name={mnt_name}")
+
+    # First test: write to source directory
     assert name not in os.listdir(src_dir)
     assert name not in os.listdir(mnt_dir)
     with open(src_name, 'w') as fh:
         fh.write('Hello, world')
+
+    print(f"Debug: File written to src_name")
+
+    start_time = time.time()
+    while time.time() - start_time < 10:  # 10 second timeout
+        if name in os.listdir(mnt_dir):
+            break
+        print(f"Debug: Waiting for file to appear... ({time.time() - start_time:.1f}s)")
+        time.sleep(0.1)
+    else:
+        pytest.fail("File did not appear in mount directory within 10 seconds")
+
     assert name in os.listdir(src_dir)
     assert name in os.listdir(mnt_dir)
-    assert os.stat(src_name) == os.stat(mnt_name)
 
+    # Compare relevant stat attributes
+    src_stat = os.stat(src_name)
+    mnt_stat = os.stat(mnt_name)
+    assert src_stat.st_mode == mnt_stat.st_mode
+    tst_inode(inode_check, mnt_stat.st_ino, src_stat.st_ino)
+    assert src_stat.st_size == mnt_stat.st_size
+    assert src_stat.st_mtime == mnt_stat.st_mtime
+
+    # Second test: write to mount directory
     name = name_generator()
     src_name = pjoin(src_dir, name)
-    mnt_name = pjoin(src_dir, name)
+    mnt_name = pjoin(mnt_dir, name)
     assert name not in os.listdir(src_dir)
     assert name not in os.listdir(mnt_dir)
     with open(mnt_name, 'w') as fh:
         fh.write('Hello, world')
     assert name in os.listdir(src_dir)
     assert name in os.listdir(mnt_dir)
-    assert os.stat(src_name) == os.stat(mnt_name)
+
+    # Compare relevant stat attributes
+    src_stat = os.stat(src_name)
+    mnt_stat = os.stat(mnt_name)
+    assert src_stat.st_mode == mnt_stat.st_mode
+    tst_inode(inode_check, mnt_stat.st_ino, src_stat.st_ino)
+    assert src_stat.st_size == mnt_stat.st_size
+    assert abs(src_stat.st_mtime - mnt_stat.st_mtime) < 0.01
 
 
 def tst_xattr(path):
